@@ -12,7 +12,9 @@ from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
 from keras.utils import to_categorical
 import numpy as np
 import time
+from sklearn.base import BaseEstimator, ClassifierMixin
 
+#region Performance Test. Compare NNRW with mainstream classifier models.
 
 def ELMClf(X, y, L = 100, verbose = False):
     '''
@@ -401,3 +403,208 @@ def PerformenceTests(func, X, y, Ls = list(range(1, 80)), N = 20):
     
     return MTacc, MVacc, MT
 
+#endregion
+
+#region ensembles
+
+def homo_stacking(X, y, create_base_estimator, Ns = [1, 2, 5], Ls = [1, 2, 5, 10, 20], repeat = 10, WITH_CONTEXT = True, xlabel = ''):
+    '''
+    create_base_estimator # a function that create base learner instanes
+    Ns = [1, 2, 5] # number of base estimators
+    Ls = [1, 2, 5, 10, 20] # base learner's specific hyper-parameter
+    repeat = 10 # run multiple times to get the mean
+    '''
+
+    pbar = tqdm(total = ITER * len(Ns) * len(Ls), position=0, leave=True) # stay on top
+
+    plt.figure(figsize = (14, 6))
+
+    for N in Ns:
+        ACCs = []
+
+        for L in Ls:        
+
+            acc = 0
+
+            for iter in range(repeat):
+
+                X_train, X_test, y_train, y_test = train_test_split(
+                    X, y
+                )
+
+                estimators = []
+                for i in range(N):
+                    estimators.append((str(i), create_base_estimator(L))) # relu is better than logistic
+
+                clf = StackingClassifier(
+                    estimators=estimators, final_estimator=LogisticRegression(), passthrough = WITH_CONTEXT
+                )
+
+                acc = acc + clf.fit(X_train, y_train).score(X_test, y_test)
+
+                pbar.update()
+
+            ACCs.append(acc / repeat)
+
+        plt.plot(Ls, ACCs, '--', label = 'N='+str(N), marker='o') # fillstyle='none'
+        # plt.scatter(Ls, ACCs, s = 50)
+
+    plt.gca().xaxis.set_major_locator(mticker.MultipleLocator(1)) # only show integer
+    plt.legend()
+    plt.xlabel('Hyper-parameter' + xlabel)
+    plt.ylabel("Classfication Accuracy")
+    plt.show()
+
+    return ACCs
+    
+def hetero_stacking(X, y, estimators, repeat = 10, WITH_CONTEXT = True):
+    '''
+    estimators: a batch of base learners. Each learner should be derived from the BaseEstimator type.    
+    '''
+    acc = 0
+    
+    for iter in range(repeat):
+
+        X_train, X_test, y_train, y_test = train_test_split(
+            X, y
+        )
+
+        clf = StackingClassifier(
+            estimators=estimators, final_estimator=LogisticRegression(), passthrough = WITH_CONTEXT
+        )
+
+        acc = acc + clf.fit(X_train, y_train).score(X_test, y_test)
+    
+    return acc / repeat
+
+class FSSE(BaseEstimator, ClassifierMixin):
+    '''
+    Feature subspace based ensemble (FSSE)
+    
+    parameters
+    ----------
+    create_base_estimator_cv : { create_elmcv_instance, create_rvflcv_instance }. 
+        a function callback that creates a cv optimized base estimator. 
+    feature_split : an integer or 'all'. The interval or "window size" for each split. 
+    meta_l1_ratios : l1_ratios to be tried for the meta-learner. We recommend > 0.5 values to achieve sparsity
+
+    Sample Code
+    -----------
+    fsse = FSSE(create_elmcv_instance, feature_split = split)
+    fsse.fit(X, y)
+    acc = fsse.evaluate(X, y) # accuracy
+    '''
+    def __init__(self, create_base_estimator_cv, feature_split = 'all', meta_l1_ratios = [0.5,0.6,0.7,0.8,0.9,1.0]):
+
+        self.create_base_estimator_cv = create_base_estimator_cv
+        self.feature_split = feature_split
+        self.meta_l1_ratios = meta_l1_ratios
+
+    def fit(self, X, y):     
+        
+        if isinstance(self.feature_split, int):
+            self.feature_split = abs(self.feature_split)
+
+        if (self.feature_split == 'all' or self.feature_split >= X.shape[1]):
+            self.feature_split = X.shape[1] 
+
+        X_train, X_test, y_train, y_test = train_test_split(X, y)
+        
+        self.base_learners = {}
+        predicts = []
+        for i in range(0, X.shape[1], self.feature_split): # TODO: padding to N*feature_split, or special treatment for the last group.
+            clfcv = self.create_base_estimator_cv().fit(X_train[:,i:i+ self.feature_split], y_train)
+            self.base_learners[(i, i + self.feature_split)] = clfcv.clf.best_estimator_ # the inner best model
+            predicts.append( clfcv.predict(X_test[:,i:i+ self.feature_split]) )
+
+        predicts = np.array(predicts).T
+        # print(predicts.shape)
+        # print(X_test.shape, y_test.shape)
+        self.meta_learner = LogisticRegressionCV(penalty = 'elasticnet',  # use elasticnet penalty to get sparse result
+                                            l1_ratios = self.meta_l1_ratios, # require at least 0.5 L1 ratio for sparsity
+                                           solver = 'saga' # only saga support elasticnet penalty
+                                           ).fit(predicts ,y_test)  
+        # print(meta_learner.l1_ratio_)
+        # plt.plot(meta_learner.coef_[0])
+        # return base_learners, meta_learner  
+
+    def predict(self, X):
+        predicts = []
+        for b in self.base_learners:
+            FS = X[:,b[0]:b[1]]
+            yhat = self.base_learners[b].predict(FS)
+            predicts.append(yhat)
+
+        predicts = np.array(predicts).T
+        return self.meta_learner.predict(predicts)
+    
+    def evaluate(self, X, y, metrics=['accuracy']):
+        y_hat = self.predict(X)        
+        return (y_hat == y).mean()
+
+    # def plot_feature_importance():
+    #    plot_feature_importance(np.abs(self.meta_learner.coef_[0]), 'FSSE FS Result') # need to include feature_importance.py
+
+    def get_important_features(self):
+        '''
+        Sample Code
+        -----------
+        biggest_fsse_fs, fs_importance = fsse.get_important_features()
+        xfsse = X_scaled[:,biggest_fsse_fs] # 前N个系数 non-zero
+        plot_feature_importance(np.abs(fs_importance), 'FSSE FS Result') # require feature_importance.py
+
+        return
+        ------
+        biggest_fsse_fs : the most important feature subset selected by meta-learner's coef
+        fs_importance : the most important features' importance
+        '''
+
+        # print(fsse.meta_learner.l1_ratio_)
+        N = np.count_nonzero(fsse.meta_learner.coef_[0])
+        
+        biggest_fsse_gs = (np.argsort(np.abs(fsse.meta_learner.coef_[0]))[-N:])[::-1] # take last N item indices and reverse (ord desc)
+        biggest_fsse_fs = []
+
+        for idx in biggest_fsse_gs:
+            biggest_fsse_fs = biggest_fsse_fs + list(range(idx*SPLIT, (idx+1)*self.feature_split))
+    
+        biggest_fsse_fs = np.array(biggest_fsse_fs)
+
+        fs_importance = []
+
+        for w in self.meta_learner.coef_[0]:
+            fs_importance = fs_importance + [w]*self.feature_split
+    
+        fs_importance = np.array(fs_importance)
+
+        return biggest_fsse_fs, fs_importance
+
+def fsse_homo_stacking(create_base_estimator_cv, split_range = range(1, 20), repeat = 3, summary = 'median'):
+
+    accs = []
+    r = split_range
+    N = repeat
+
+    for split in r:
+
+        for i in range(N):
+
+            accs_repeat = []
+
+            fsse = FSSE(create_base_estimator_cv, feature_split = split)
+            fsse.fit(X_scaled, y)
+            accs_repeat.append(fsse.evaluate(X_scaled, y))
+
+        if summary == 'median':
+            accs.append( np.median(accs_repeat) )
+        else:
+            accs.append( np.mean(accs_repeat) ) # otherwise, mean
+
+    plt.figure(figsize = (12,3))
+    plt.scatter(r,accs)
+    plt.plot(r, accs)
+    plt.show()
+    
+    return accs
+
+#endregion
